# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2018-08-21 10:42:16
from __future__ import unicode_literals
import datetime

version = "2.0.6"
githash = "5e30a46dd0146a241f5c7ca8de0cc0144455b2e1"


major = 2
minor = 0
bugfix = 6

release = True
timestamp = datetime.datetime(2018, 8, 21, 10, 42, 16)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
