# The ``astropy_helpers.extern`` sub-module includes modules developed elsewhere
# that are bundled here for convenience. At the moment, this consists of the
# following two sphinx extensions:
#
# * `numpydoc <https://github.com/numpy/numpydoc>`_, a Sphinx extension
#   developed as part of the Numpy project. This is used to parse docstrings
#   in Numpy format
#
# * `sphinx-automodapi <https://github.com/astropy/sphinx-automodapi>`_, a Sphinx
#   developed as part of the Astropy project. This used to be developed directly
#   in ``astropy-helpers`` but is now a standalone package.
